# SPDX-License-Identifier: BSD-2-Clause

from .pmod import (
	PmodDualHBridgeType6Resource, PmodGPIOType1Resource,
	PmodHBridgeType5Resource, PmodSPIType2AResource,
	PmodSPIType2Resource, PmodUARTType3Resource,
	PmodUARTType4AResource, PmodUARTType4Resource
)

__all__ = (
	'PmodDualHBridgeType6Resource',
	'PmodGPIOType1Resource',
	'PmodHBridgeType5Resource',
	'PmodSPIType2AResource',
	'PmodSPIType2Resource',
	'PmodUARTType3Resource',
	'PmodUARTType4AResource',
	'PmodUARTType4Resource',
)
