# SPDX-License-Identifier: BSD-2-Clause

from .dsl  import (
	Attrs, Clock, Connector, DiffPairs, DiffPairsN, Pins, PinsN,
	Resource, Subsignal
)
from .plat import Platform, TemplatedPlatform
from .res  import ResourceError

__all__ = (
	'Attrs',
	'Clock',
	'Connector',
	'DiffPairs',
	'DiffPairsN',
	'Pins',
	'PinsN',
	'Platform',
	'Resource',
	'ResourceError',
	'Subsignal',
	'TemplatedPlatform',
)
